from sympy import asin, acos, atan, acsc, asec, acot, log, sqrt
from sympy import asinh, acosh, atanh, acsch, asech, acoth
from sympy import Xor, And, Or, Rational, Abs
from sympy import (
    Equality,
    Unequality,
    LessThan,
    StrictLessThan,
    GreaterThan,
    StrictGreaterThan,
)
from sympy import Symbol, diff

import numbers


def tree_node(op, args):
    args = [expr(a) for a in args]
    return {"type": "Tree", "op": op, "args": args}


def operation(func):
    op = str(func)
    if func == sqrt:
        op = "root"
    elif func == log: 
        op = "ln"   # this is confusing but sympy uses `log` for natural logarithm
    elif func == Abs:
        op = "abs"
    elif func == asin:
        op = "arcsin"
    elif func == acos:
        op = "arccos"
    elif func == atan:
        op = "arctan"
    elif func == asinh:
        op = "arcsinh"
    elif func == acosh:
        op = "arccosh"
    elif func == atanh:
        op = "arctanh"
        
    return op


def process_mul(y):
    assert(y.is_Mul)
    if len(y.args) == 2 and y.args[1].is_Pow and y.args[1].args[1] == -1:
        return tree_node("divide", [y.args[0], y.args[1].args[0]])
    else:
        return tree_node("times", y.args)


def process_pow(y):
    assert(y.is_Pow)
    power = y.args[1]
    arg = y.args[0]

    if power == 2:
        return tree_node("square", [arg])
    elif power == 3:
        return tree_node("cube", [arg])
    elif power == -1:
        return tree_node("recip", [arg])
    elif power == -2:
        return tree_node("recip", [arg**2])
    elif power == -3:
        return tree_node("recip", [arg**3])
    elif power == Rational(1, 2):
        return tree_node("root", [arg])
    elif power == Rational(3, 2):
        return tree_node("root", [arg**3])
    else:
        return tree_node("power", y.args)


def tree(y):
    if y.is_Add:
        return tree_node("plus", y.args)
    elif y.is_Mul:        
        return process_mul(y)
    elif y.is_Pow:
        return process_pow(y)
    elif y.is_Function:
        return tree_node(operation(y.func), y.args)        
    else:
        raise ValueError("unreognized tree type")


def relational(y):
    f = y.func
    op = ""

    if f == LessThan:
        op = "lt"
    elif f == StrictLessThan:
        op = "leq"
    elif f == GreaterThan:
        op = "gt"
    elif f == StrictGreaterThan:
        op = "geq"
    elif f == Equality:
        op = "eq"
    elif f == Unequality:
        op = "neq"
    else:
        raise ValueError("unrecognized relational operator")

    return tree_node(op, y.args)


def boolean(y):
    f = y.func
    op = ""

    if f == And:
        op = "and"
    elif f == Or:
        op = "or"
    elif f == Xor:
        op = "xor"
    else:
        raise ValueError("unrecognized boolean operator")

    return tree_node(op, y.args)


def piecewise(y):
    cond = y.args[0][1]
    x1 = y.args[0][0]

    if len(y.args) == 1:
        return expr(x1)
    if len(y.args) == 2:
        x2 = y.args[1][0]
    else:
        x2 = piecewise(*y.args[1:])

    return tree_node("ifelse", [cond, x1, x2])


def var(sym, val=0.0):
    return {"name": sym.name, "val": float(val)}


def expr(y):
    if y.is_Number or isinstance(y, numbers.Number):
        return {"type": "Const", "val": float(y)}
    elif y.is_Symbol:
        return {"type": "Var", "name": y.name}
    elif y.is_Relational:
        return relational(y)
    elif y.is_Boolean:
        return boolean(y)
    elif y.is_Piecewise:
        return piecewise(y)
    else:
        return tree(y)


def equation(lhs, rhs):
    return {"lhs": lhs, "rhs": rhs}


def ode(y):
    return {
        "type": "Tree",
        "op": "Differential",
        "args": [{"type": "Var", "name": y.name}],
    }


def model(states, eqs, params=None, obs=None):
    if not isinstance(states, list):
        states = [states]

    if not isinstance(eqs, list):
        eqs = [eqs]

    if params is None:
        params = []
        
    if obs is None:
        obs = [Symbol(f"${i}") for i in range(len(eqs))]

    d = {
        "iv": var(Symbol("$_")),
        "params": [var(x) for x in params],
        "states": [var(x) for x in states],
        "algs": [],
        "odes": [],
        "obs": [
            equation(expr(lhs), expr(rhs)) for (lhs, rhs) in zip(obs, eqs)
        ],
    }

    return d


def model_ode(iv, states, odes, params=None):
    try:
        states = list(states)
    except TypeError:
        states = [states]

    try:
        odes = list(odes)
    except TypeError:
        odes = [odes]

    assert len(states) == len(odes)

    if params is None:
        params = []

    d = {
        "iv": var(iv),
        "params": [var(x) for x in params],
        "states": [var(x) for x in states],
        "algs": [],
        "odes": [equation(ode(lhs), expr(rhs)) for (lhs, rhs) in zip(states, odes)],
        "obs": [],
    }

    return d
    
 
def model_jac(iv, states, odes, params=None):
    try:
        states = list(states)
    except TypeError:
        states = [states]

    try:
        odes = list(odes)
    except TypeError:
        odes = [odes]

    assert len(states) == len(odes)
    
    n = len(states)
    eqs = []
    
    for i in range(n):
        for j in range(n):
            df = diff(odes[i], states[j])
            eqs.append(df)

    if params is None:
        params = []

    d = {
        "iv": var(iv),
        "params": [var(x) for x in params],
        "states": [var(x) for x in states],
        "algs": [],
        "odes": [],
        "obs": [equation(expr(Symbol(f"${i}")), expr(rhs)) for (i, rhs) in enumerate(eqs)],
    }

    return d



import os
import sys
import ctypes
import platform

def find_dll(substr):
    files = os.listdir(os.path.dirname(__file__))
    matches = list(filter(lambda s: s.find(substr) >= 0, files))
    if len(matches) == 0:
        return None
    else:
        return matches[0]


dll_name = None

if sys.platform == "linux" and platform.machine() == "x86_64":
    dll_name = find_dll("x86_64-linux")
if sys.platform == "linux" and platform.machine() == "aarch64":
    dll_name = find_dll("aarch64-linux")
if sys.platform == "darwin":
    dll_name = find_dll("darwin")
elif sys.platform == "win32":
    dll_name = find_dll("win_amd64")

if dll_name is None:
    raise ValueError("unsupported platform (consider installing symfit from conda-forge as `conda install -c conda-forge symjit`)")

# print(dll_name)

dll_path = os.path.join(os.path.dirname(__file__), dll_name)
dll = ctypes.CDLL(dll_path)


class Engine:
    def __init__(self):
        self._info = dll.info
        self._info.argtypes = []
        self._info.restype = ctypes.c_char_p

        self._check_status = dll.check_status
        self._check_status.argtypes = [ctypes.c_void_p]
        self._check_status.restype = ctypes.c_char_p

        self._count_states = dll.count_states
        self._count_states.argtypes = [ctypes.c_void_p]
        self._count_states.restype = ctypes.c_size_t

        self._count_params = dll.count_params
        self._count_params.argtypes = [ctypes.c_void_p]
        self._count_params.restype = ctypes.c_size_t

        self._count_obs = dll.count_obs
        self._count_obs.argtypes = [ctypes.c_void_p]
        self._count_obs.restype = ctypes.c_size_t

        self._count_diffs = dll.count_diffs
        self._count_diffs.argtypes = [ctypes.c_void_p]
        self._count_diffs.restype = ctypes.c_size_t

        self._run = dll.run
        self._run.argtypes = [
            ctypes.c_void_p,  # handle
            ctypes.POINTER(ctypes.c_double),  # du
            ctypes.POINTER(ctypes.c_double),  # u
            ctypes.c_size_t,  # ns
            ctypes.POINTER(ctypes.c_double),  # p
            ctypes.c_size_t,  # np
            ctypes.c_double,  # t
        ]
        self._run.restype = ctypes.c_bool

        self._execute = dll.execute
        self._execute.argtypes = [
            ctypes.c_void_p,  # handle
            ctypes.c_double,  # t
        ]
        self._execute.restype = ctypes.c_bool
        
        self._execute_vectorized = dll.execute_vectorized
        self._execute_vectorized.argtypes = [
            ctypes.c_void_p,  # handle
            ctypes.POINTER(ctypes.c_double),    # buf
            ctypes.c_size_t                     # n
        ]
        self._execute_vectorized.restype = ctypes.c_bool

        self._fill_u0 = dll.fill_u0
        self._fill_u0.argtypes = [
            ctypes.c_void_p,  # handle
            ctypes.POINTER(ctypes.c_double),  # u0
            ctypes.c_size_t,  # ns
        ]
        self._fill_u0.restype = ctypes.c_bool

        self._fill_p = dll.fill_p
        self._fill_p.argtypes = [
            ctypes.c_void_p,  # handle
            ctypes.POINTER(ctypes.c_double),  # p
            ctypes.c_size_t,  # np
        ]
        self._fill_p.restype = ctypes.c_bool

        self._compile = dll.compile
        self._compile.argtypes = [ctypes.c_char_p, ctypes.c_char_p, ctypes.c_bool]
        self._compile.restype = ctypes.c_void_p

        self._dump = dll.dump
        self._dump.argtypes = [ctypes.c_void_p, ctypes.c_char_p, ctypes.c_char_p]
        self._dump.restype = ctypes.c_bool

        self._finalize = dll.finalize
        self._finalize.argtypes = [ctypes.c_void_p]
        self._finalize.restype = None

        self._ptr_states = dll.ptr_states
        self._ptr_states.argtypes = [ctypes.c_void_p]
        self._ptr_states.restype = ctypes.POINTER(ctypes.c_double)

        self._ptr_params = dll.ptr_params
        self._ptr_params.argtypes = [ctypes.c_void_p]
        self._ptr_params.restype = ctypes.POINTER(ctypes.c_double)

        self._ptr_obs = dll.ptr_obs
        self._ptr_obs.argtypes = [ctypes.c_void_p]
        self._ptr_obs.restype = ctypes.POINTER(ctypes.c_double)

        self._ptr_diffs = dll.ptr_diffs
        self._ptr_diffs.argtypes = [ctypes.c_void_p]
        self._ptr_diffs.restype = ctypes.POINTER(ctypes.c_double)

    def info(self):
        return self._info()




import json
import numpy as np
import numbers
import ctypes

from . import engine
from . import structure

lib = engine.Engine()   # interface to the rust codegen engine


def from_raw_parts(ptr, count):
    return np.ctypeslib.as_array(ptr, shape=(count,))


class BaseFunc:
    def __init__(self, model, ty="native", use_simd=True):
        self.p = lib._compile(model.encode("utf-8"), ty.encode("utf8"), use_simd)
        status = lib._check_status(self.p)
        if status != b"Success":
            raise ValueError(status.decode())
        self.populate()
        self.model = model  # for debugging

    def __del__(self):
        # lib._finalize(self.p)   
        pass  
        
    def get_u0(self):
        u0 = np.zeros(self.count_states, dtype="double")
        lib._fill_u0(self.p, np.ctypeslib.as_ctypes(u0), self.count_states)
        return u0

    def get_p(self):
        p = np.zeros(self.count_params, dtype="double")
        lib._fill_p(self.p, np.ctypeslib.as_ctypes(p), self.count_params)
        return p

    def populate(self):
        self.count_states = lib._count_states(self.p)
        self.count_params = lib._count_params(self.p)
        self.count_obs = lib._count_obs(self.p)
        self.count_diffs = lib._count_diffs(self.p)

        self._states = from_raw_parts(lib._ptr_states(self.p), self.count_states)
        self._params = from_raw_parts(lib._ptr_params(self.p), self.count_params)
        self._obs = from_raw_parts(lib._ptr_obs(self.p), self.count_obs)
        self._diffs = from_raw_parts(lib._ptr_diffs(self.p), self.count_diffs)
        
    def dump(self, name, what="scalar"):
        if not lib._dump(self.p, name.encode("utf-8"), what.encode("utf-8")):
            print("cannot dump the requested code")
        

class Func(BaseFunc):
    def __init__(self, model, ty="native", use_simd=True):
        super().__init__(model, ty=ty, use_simd=use_simd)

    def __call__(self, *args):
        if len(args) > self.count_states:
            p = np.array(args[self.count_states:], dtype="double")
            self._params[:] = p
    
        if isinstance(args[0], numbers.Number):
            u = np.array(args[:self.count_states], dtype="double")
            self._states[:] = u
            status = lib._execute(self.p, 0.0)

            if not status:
                raise ValueError("cannot execute the model")

            return self._obs.copy()
        else:
            return self.call_vectorized(*args)
            
    def call_vectorized(self, *args):
        assert(len(args) >= self.count_states)
        shape = args[0].shape
        n = args[0].size
        h = max(self.count_states, self.count_obs)
        buf = np.zeros((h, n), dtype="double")

        for i in range(self.count_states):
            assert(args[i].shape == shape)
            buf[i,:] = args[i].ravel()            
        
        ptr = buf.ctypes.data_as(ctypes.POINTER(ctypes.c_double))
        status = lib._execute_vectorized(self.p, ptr, n)
        
        if not status:
            raise ValueError("cannot execute the model")
        
        res = []
        for i in range(self.count_obs):            
            y = buf[i,:].reshape(shape)
            res.append(y)
            
        return res


class OdeFunc(BaseFunc):
    def __init__(self, model, ty="native", use_simd=False):
        super().__init__(model, ty=ty, use_simd=use_simd)

    def __call__(self, t, y, *args):
        y = np.array(y, dtype="double")
        self._states[:] = y

        if len(args) > 0:
            p = np.array(args, dtype="double")
            self._params[:] = p

        status = lib._execute(self.p, t)

        if not status:
            raise ValueError("cannot execute the model")

        return self._diffs.copy()


class JacFunc(BaseFunc):
    def __init__(self, model, ty="native", use_simd=False):
        super().__init__(model, ty=ty, use_simd=use_simd)

    def __call__(self, t, y, *args):
        y = np.array(y, dtype="double")
        self._states[:] = y

        if len(args) > 0:
            p = np.array(args, dtype="double")
            self._params[:] = p

        status = lib._execute(self.p, t)

        if not status:
            raise ValueError("cannot execute the model")

        jac = self._obs.copy()        
        return jac.reshape((self.count_states, self.count_states))


def compile_func(states, eqs, params=None, obs=None, ty="native", use_simd=True):
    """Compile a list of symbolic expression into an executable form.
    compile_func tries to mimic sympy lambdify, but instead of generating
    a standard python funciton, it returns a callable (Func object) that
    is a thin wrapper over compiled machine-code.
    
    Parameters
    ==========
    
    states: a single symbol or a list/tuple of symbols.
    eqs: a single symbolic expression or a list/tuple of symbolic expressions.
    params (optional): a list/tuple of additional symbols as parameters to the model.
    ty: target architecture ("amd", "arm", "bytecode", or "native").
    obs (optional): a list of symbols to name equations. If obs is not None, its length should 
        be the same as eqs.
    use_simd (default True): generates SIMD code for vectorized operations. 
        Currently only AVX on X64 systems is supported.
    
    ==> returns a Func object, is a callable object `f` with signature `f(x_1,...,x_n,p_1,...,p_m)`, 
        where `x`s are the state variables and `p`s are the parameters.
    
    >>> import numpy as np
    >>> from symjit import compile_func
    >>> from sympy import symbols
    
    >>> x, y = symbols('x y')
    >>> f = compile_func([x, y], [x+y, x*y])
    >>> assert(np.all(f(3, 5) == [8., 15.]))
    """
    model = structure.model(states, eqs, params=params, obs=obs)
    return Func(json.dumps(model), ty=ty, use_simd=use_simd)


def compile_ode(iv, states, odes, params=None, ty="native", use_simd=False):
    """Compile a symbolic ODE model into an executable form suitable for 
    passung to scipy.integrate.solve_ivp.    
    
    Parameters
    ==========
    
    iv: a single symbol, the independent variable.
    states: a single symbol or a list/tuple of symbols.
    odes: a single symbolic expression or a list/tuple of symbolic expressions,
        representing the derivative of the state with respect to iv.
    params (optional): a list/tuple of additional symbols as parameters to the model
    ty: target architecture ("amd", "arm", "bytecode", or "native").
    use_simd (default False): generates SIMD code for vectorized operations. 
        Currently only AVX on X64 systems is supported.
    
    invariant => len(states) == len(odes)
    
    ==> returns an OdeFunc object, is a callable object `f` with signature `f(t,y,p0,p1,...)`, 
        where `t` is the value of the independent variable, `y` is the state (an array of 
        state variables), and `p`s are the parameters. 
    
    >>> import scipy.integrate
    >>> import numpy as np
    >>> from sympy import symbols
    >>> from symjit import compile_ode
    
    >>> t, x, y = symbols('t x y')
    >>> f = compile_ode(t, (x, y), (y, -x))
    >>> t_eval=np.arange(0, 10, 0.01)
    >>> sol = scipy.integrate.solve_ivp(f, (0, 10), (0.0, 1.0), t_eval=t_eval)

    >>> np.testing.assert_allclose(sol.y[0,:], np.sin(t_eval), atol=0.005)
    """
    model = structure.model_ode(iv, states, odes, params)
    return OdeFunc(json.dumps(model), ty=ty, use_simd=use_simd)
    
def compile_jac(iv, states, odes, params=None, ty="native", use_simd=False):
    """Genenrates and compiles Jacobian for an ODE system.
        It accepts the same arguments as `compile_ode`.
        
    ===> returns an OdeFunc object that has the same signature as 
        the results of `compile_ode`, i.e., `f(t,y,p0,p1,...)`. 
        However, it returns a n-by-n Jacobian matrix, where n is
        the number of state variables. 
    """
    model = structure.model_jac(iv, states, odes, params)
    return JacFunc(json.dumps(model), ty=ty, use_simd=use_simd)

def compile_json(model, ty="native", use_simd=True):
    return OdeFunc(model, ty=ty, use_simd=use_simd)
